/*
THE COMPUTER CODE CONTAINED HEREIN IS THE SOLE PROPERTY OF PARALLAX
SOFTWARE CORPORATION ("PARALLAX").  PARALLAX, IN DISTRIBUTING THE CODE TO
END-USERS, AND SUBJECT TO ALL OF THE TERMS AND CONDITIONS HEREIN, GRANTS A
ROYALTY-FREE, PERPETUAL LICENSE TO SUCH END-USERS FOR USE BY SUCH END-USERS
IN USING, DISPLAYING,  AND CREATING DERIVATIVE WORKS THEREOF, SO LONG AS
SUCH USE, DISPLAY OR CREATION IS FOR NON-COMMERCIAL, ROYALTY OR REVENUE
FREE PURPOSES.  IN NO EVENT SHALL THE END-USER USE THE COMPUTER CODE
CONTAINED HEREIN FOR REVENUE-BEARING PURPOSES.  THE END-USER UNDERSTANDS
AND AGREES TO THE TERMS HEREIN AND ACCEPTS THE SAME BY USE OF THIS FILE.  
COPYRIGHT 1993-1998 PARALLAX SOFTWARE CORPORATION.  ALL RIGHTS RESERVED.
*/

/* WHS: I grabbed the version changes list from the PC version's digi.c as there were none here and some info is still relevant (and possibly interesting). */


/*
 * $Source: f:/miner/source/main/rcs/digi.c $
 * $Revision: 2.5 $
 * $Author: john $
 * $Date: 1996/01/05 16:51:51 $
 * 
 * Routines to access digital sound hardware
 * 
 * $Log: digi.c $
 * Revision 2.5  1996/01/05  16:51:51  john
 * Made the midi handler lock down to +4K.
 * 
 * Revision 2.4  1996/01/05  16:46:54  john
 * Made code lock down midicallback +4K.
 * 
 * Revision 2.3  1995/05/26  16:16:03  john
 * Split SATURN into define's for requiring cd, using cd, etc.
 * Also started adding all the Rockwell stuff.
 * 
 * Revision 2.2  1995/03/30  16:36:04  mike
 * text localization.
 * 
 * Revision 2.1  1995/03/15  11:41:23  john
 * Better Saturn CD-ROM support.
 * 
 * Revision 2.0  1995/02/27  11:29:15  john
 * New version 2.0, which has no anonymous unions, builds with
 * Watcom 10.0, and doesn't require parsing BITMAPS.TBL.
 * 
 * Revision 1.172  1995/02/15  09:57:53  john
 * Fixed bug with loading new banks while song is playing.
 * 
 * Revision 1.171  1995/02/14  16:36:26  john
 * Fixed bug with countdown voice getting cut off.
 * 
 * Revision 1.170  1995/02/13  20:34:51  john
 * Lintized
 * 
 * Revision 1.169  1995/02/13  15:18:21  john
 * Added code to reset sounds.
 * 
 * Revision 1.168  1995/02/13  12:53:11  john
 * Fixed bug with demos recording wrong volumes.
 * 
 * Revision 1.167  1995/02/11  15:04:29  john
 * Made FM files load as if you have FM card.
 * 
 * Revision 1.166  1995/02/11  12:41:35  john
 * Added new song method, with FM bank switching..
 * 
 * Revision 1.165  1995/02/10  16:24:58  john
 * MAde previous change only reverse for RAP10.
 * 
 * Revision 1.164  1995/02/10  13:47:37  john
 * Made digi init before midi to maybe fix RAP10 bug.
 * 
 * Revision 1.163  1995/02/08  21:05:22  john
 * Added code that loads patches for every FM song.
 * 
 * Revision 1.162  1995/02/03  17:17:51  john
 * Made digi lowmem default to off.
 * 
 * Revision 1.161  1995/02/03  17:08:22  john
 * Changed sound stuff to allow low memory usage.
 * Also, changed so that Sounds isn't an array of digi_sounds, it
 * is a ubyte pointing into GameSounds, this way the digi.c code that
 * locks sounds won't accidentally unlock a sound that is already playing, but
 * since it's Sounds[soundno] is different, it would erroneously be unlocked.
 * 
 * Revision 1.160  1995/02/01  22:37:06  john
 * Reduced sound travel distance to 1.25 times original.
 * 
 * Revision 1.159  1995/02/01  22:20:41  john
 * Added digi_is_sound_playing.
 * 
 * Revision 1.158  1995/01/28  15:56:56  john
 * Made sounds carry 1.5 times farther.
 * 
 * Revision 1.157  1995/01/27  17:17:09  john
 * Made max sounds relate better to MAX_SOUND_OBJECTS
 * 
 * Revision 1.156  1995/01/25  12:18:06  john
 * Fixed bug with not closing MIDI files when midi volume is 0.
 * 
 * Revision 1.155  1995/01/24  17:52:17  john
 * MAde midi music stop playing when volume is 0.
 * 
 * Revision 1.154  1995/01/21  21:22:31  mike
 * Correct bogus error message.
 * 
 * Revision 1.153  1995/01/17  14:53:38  john
 * IFDEF'D out digital drums.
 * 
 * Revision 1.152  1995/01/11  16:26:50  john
 * Restored MIDI pausing to actually setting volume to 0,.
 * 
 * Revision 1.151  1995/01/10  16:38:46  john
 * Made MIDI songs pause, not lower volume when pausing, and
 * had to fix an HMI bug by clearing the TRACKINFO array or something.
 * 
 * Revision 1.150  1995/01/05  19:46:27  john
 * Added code to reset the midi tracks to use the ones 
 * in the midi file before each song is played.
 * 
 * Revision 1.149  1994/12/21  15:08:59  matt
 * Bumped MAX_SOUND_OBJECTS back up to match v1.00 of shareware
 * 
 * Revision 1.148  1994/12/20  18:22:54  john
 * Added code to support non-looping songs, and put
 * it in for endlevel and credits.
 * 
 * Revision 1.147  1994/12/19  17:58:19  john
 * Changed Assert for too many sounds to gracefully exit.
 * 
 * Revision 1.146  1994/12/14  16:03:27  john
 * Made the digi/midi deinit in reverse order for anything
 * other than GUS.
 * 
 * Revision 1.145  1994/12/14  14:51:06  john
 * Added assert.
 * 
 * Revision 1.144  1994/12/14  12:14:40  john
 * Relplaced a bunch of (a*b)/c with fixmuldiv
 * to get rid of a overflow bug mike found.
 * 
 * Revision 1.143  1994/12/13  17:30:16  john
 * Made the timer rate be changed right after initializing it.
 * 
 * Revision 1.142  1994/12/13  14:08:35  john
 * Made Pause key set midi volume to 0.
 * Made digi_init set midi volume to proper level.
 * 
 * Revision 1.141  1994/12/13  12:42:31  jasen
 * Fixed sound priority bugs... -john
 * 
 * Revision 1.140  1994/12/13  12:11:38  john
 * Added debugging code.
 * 
 * Revision 1.139  1994/12/13  11:45:19  john
 * Disabled interrupts around the midisetvolume because awe32
 * hangs if you don't.
 * 
 * Revision 1.138  1994/12/13  11:33:45  john
 * MAde so that sounds with volumes > f1_0 don't cut off.
 * 
 * Revision 1.137  1994/12/13  02:24:29  matt
 * digi_init() now doesn't return error when no sound card
 * 
 * Revision 1.136  1994/12/13  00:46:27  john
 * Split digi and midi volume into 2 seperate functions.
 * 
 * Revision 1.135  1994/12/12  22:19:20  john
 * Made general midi versions of files load...
 * .hmq instead of .hmp.
 * 
 * Revision 1.134  1994/12/12  21:32:49  john
 * Made volume work better by making sure volumes are valid
 * and set correctly at program startup.
 * 
 * Revision 1.133  1994/12/12  20:52:35  john
 * Fixed bug with pause calling set mastervolume to 0.
 * 
 * Revision 1.132  1994/12/12  20:39:52  john
 * Changed so that instead of using MasterVolume for 
 * digital sounds, I just scale the volume I play the
 * sound by.
 * 
 * Revision 1.131  1994/12/12  13:58:21  john
 * MAde -nomusic work.
 * Fixed GUS hang at exit by deinitializing digi before midi.
 * 
 * Revision 1.130  1994/12/11  23:29:39  john
 * *** empty log message ***
 * 
 * Revision 1.129  1994/12/11  23:18:07  john
 * Added -nomusic.
 * Added RealFrameTime.
 * Put in a pause when sound initialization error.
 * Made controlcen countdown and framerate use RealFrameTime.
 * 
 * Revision 1.128  1994/12/11  13:35:26  john
 * Let sounds play louder than F1_0.  btw, 4*f1_0 is maxed.
 * 
 * Revision 1.127  1994/12/11  00:43:45  john
 * Fixed bug with sounds taking over channels that sound objects own.
 * 
 * Revision 1.126  1994/12/10  20:35:03  john
 * Added digi_kill_sound_linked_to_object.
 * 
 * Revision 1.125  1994/12/10  15:44:20  john
 * Added max_distance passing for sound objects.
 * 
 * Revision 1.124  1994/12/09  20:16:37  john
 * Made it so that when Mike's AI code passes bogus values, the
 * digi code will saturate them.
 * 
 * Revision 1.123  1994/12/09  17:07:52  john
 * Fixed abrupt fan cutoff.
 * 
 * Revision 1.122  1994/12/08  17:19:14  yuan
 * Cfiling stuff.
 * 
 * Revision 1.121  1994/12/06  19:24:08  john
 * CLosed HMI timer after making our not call it.
 * 
 * Revision 1.120  1994/12/06  19:00:42  john
 * Moved digi_reset_sounds to reset instead of digi_close.
 * 
 * Revision 1.119  1994/12/06  18:23:48  matt
 * Don't pause midi songs, becuase it can lock up
 * 
 * Revision 1.118  1994/12/06  16:07:09  john
 * MAde the gus pats only load if using midi==midi_gus.
 * 
 * Revision 1.117  1994/12/06  10:17:07  john
 * MAde digi_close call reset_sounds.
 * 
 * Revision 1.116  1994/12/05  23:36:50  john
 * Took out lock down of GETDS, because, I, john, was wrong, and it
 * was me, not the HMI people, that was calling GETDS in an interrupt.
 * I un-formally apologize to HMI.
 * 
 * Revision 1.115  1994/12/05  22:50:09  john
 * Put in code to lock down GETDS, because the HMI
 * people don't lock it down, even tho they use it 
 * in their timerhandler interrupt.
 * 
 * Revision 1.114  1994/12/05  18:54:09  john
 * *** empty log message ***
 * 
 * Revision 1.113  1994/12/05  18:52:11  john
 * Locked down the TrackMap array, since SOS references it in an interupt.
 * Added -noloadpats command line option.
 * 
 * Revision 1.112  1994/12/05  16:55:45  john
 * Made so that digi_init_System is  called before 
 * the timer system calls the timer handler.
 * 
 * Revision 1.111  1994/12/05  12:56:19  john
 * Made the intruments patches only load for FM devices.
 * 
 * Revision 1.110  1994/12/05  12:17:16  john
 * Added code that locks/unlocks digital sounds on demand.
 * 
 * Revision 1.109  1994/12/04  14:30:10  john
 * Added hooks for music..
 * 
 * Revision 1.108  1994/12/02  13:58:27  matt
 * Put in Int3()s when try to play a non-existant sound
 * 
 * Revision 1.107  1994/12/02  10:35:58  john
 * Took out loadpats.c
 * 
 * Revision 1.106  1994/12/01  02:22:33  john
 * Incorporated LOADPATS.EXE into our code.
 * 
 * Revision 1.105  1994/12/01  00:59:51  john
 * Fixed some pot. bugs with closing.
 * 
 * Revision 1.104  1994/11/30  23:54:40  rob
 * Tweaked some volume params due to a fix in find_connected_distance.
 * 
 * Revision 1.103  1994/11/30  19:36:44  john
 * Made Gravis Ultrasound work again.  Made the scores blink
 * at a constant rate.  Revamped the newmenu background storage,
 * which hopefully fixed some bugs.  Made menus in ame not pause
 * sound, except for the pause key.               ^== Game!
 * 
 * Revision 1.102  1994/11/30  15:14:25  rob
 * Removed unused include file..
 * 
 * Revision 1.101  1994/11/30  15:08:45  john
 * Changed some open/close stuff with sound.
 * 
 * Revision 1.100  1994/11/29  13:35:41  john
 * Test code.
 * 
 * Revision 1.99  1994/11/29  03:46:32  john
 * Added joystick sensitivity; Added sound channels to detail menu.  Removed -maxchannels
 * command line arg.
 * 
 * Revision 1.98  1994/11/28  18:37:59  john
 * Made sample play once work.
 * 
 * Revision 1.97  1994/11/28  18:35:19  john
 * Made the digi_max_channels cut of an old sound instead of
 * not playing a new sound.
 * 
 * Revision 1.96  1994/11/28  01:32:02  mike
 * use quick form of matrix function.
 * 
 * Revision 1.95  1994/11/27  23:12:14  matt
 * Made changes for new mprintf calling convention
 * 
 * Revision 1.94  1994/11/22  17:13:31  john
 * Fixed bug with the digital drums hanging.
 * 
 * Revision 1.93  1994/11/21  19:09:21  john
 * Made so if digi and midi both 0, same as -nosound.
 * 
 * Revision 1.92  1994/11/21  16:46:46  john
 * Limited digital channels to 10.
 * 
 * Revision 1.91  1994/11/21  16:28:42  john
 * Fixed bug with digi_close hanging.
 * 
 * Revision 1.90  1994/11/21  15:40:28  john
 * Locked down the instrument data.
 * 
 * Revision 1.89  1994/11/21  14:43:55  john
 * Fixed some bugs with setting volumes even when -nosound was used. Duh!
 * 
 * Revision 1.88  1994/11/21  11:55:26  john
 * Fixed some sound pausing in menus bugs.
 * 
 * Revision 1.87  1994/11/21  11:02:21  john
 * Made fan sounds be 1/2 the volume so that they
 * don'
 * don't carry as far.
 * 
 * Revision 1.86  1994/11/20  17:47:51  john
 * Fixed a potential bug with sound initializing.
 * 
 * Revision 1.85  1994/11/20  17:29:07  john
 * Fixed bug with page fault during digi_close.
 * 
 * Revision 1.84  1994/11/19  15:19:24  mike
 * rip out unused code and data.
 * 
 * Revision 1.83  1994/11/16  23:38:33  mike
 * new improved boss teleportation behavior.
 * 
 * Revision 1.82  1994/11/14  18:12:46  john
 * Took out some sound objects stuff.
 * 
 */

#include <stdio.h>
#include <stdlib.h>

#include "inferno.h"
#include "texmap.h"		// For Lighting_on
#include "songs.h"
#include "screens.h"
#include "game.h"
#include "player.h"
#include "cfile.h"
#include "segpoint.h"
#include "ai.h"
#include "piggy.h"
#include "wall.h"
#include "joy.h"
//#include "halestrm.h"
#include "byteswap.h"
#include "mem.h"
#include "sounds.h"
#include "mono.h"
#include "error.h"
#include "kconfig.h"
#include "redbook.h"
#include "newdemo.h"
#include "macsys.h"

//#define DEBUG /* uncomment to get debug output for the sounds */
#include "debug.h"
#ifdef USE_SOUND
#  include <gsi/gsi.h>
#endif


#define _DIGI_SAMPLE_FLAGS (_VOLUME | _PANNING )
#define _DIGI_MAX_VOLUME			256
#define MAX_SOUND_CHANNELS			24
#ifdef MAC_SHAREWARE
#  define MAX_MUSIC_CHANNELS			6
Handle Song_handle = NULL;
Handle Midi_handle = NULL;
#endif
#ifdef MAC_REGISTERED_15
#  define MAX_MUSIC_CHANNELS			0
#endif
#ifdef PC_REGISTERED_15
#  define MAX_MUSIC_CHANNELS			16
#endif

int digi_lomem = 0;
int master_save;
int digi_initialized = 0;
ubyte digi_paused = 0;
ubyte digi_cd_started = 0;			// have we started a disc that is not the Descent cd?
int midi_volume = 255;
int digi_volume = _DIGI_MAX_VOLUME;
int digi_midi_song_playing = 0;
int digi_last_midi_song = 0;
int digi_last_midi_song_loop = 0;
//WHS: added following, what should other consts be?
#define   Q_NORMAL 0
#define   Q_HIGH   1
#define SOUND_END -1 //ok? WHS
int digi_sound_quality =  Q_NORMAL;

int digi_max_channels = MAX_SOUND_CHANNELS;
ushort num_sounds;

extern ubyte Config_master_volume;

// sound object stuff -- used for fans and the boss only??
#define SOF_USED			1 		// Set if this sample is used
#define SOF_PLAYING			2		// Set if this sample is playing on a channel
#define SOF_LINK_TO_OBJ		4		// Sound is linked to a moving object. If object dies, then finishes play and quits.
#define SOF_LINK_TO_POS		8		// Sound is linked to segment, pos
#define SOF_PLAY_FOREVER	16		// Play forever (or until level is stopped), otherwise plays once

#define SOUND_PLAYING			1
#define SOUND_OBJECT_PLAYING	2

//Only used in mac version, PC version uses the song_info struct (main/songs.h)
typedef struct song_resource {
	short		midi_id;
	ubyte		lead_inst;
	ubyte		buffer_ahead;
	ushort		tempo;
	ushort		pitch_shift;
	ubyte		sound_voices;
	ubyte		max_notes;
	ushort		norm_voices;
} song_resource;

typedef struct sound_object {
	short			signature;		// A unique signature to this sound
	ubyte			flags;			// Used to tell if this slot is used and/or currently playing, and how long.
	fix				max_volume;		// Max volume that this sound is playing at
	fix				max_distance;	// The max distance that this sound can be heard at...
	int				volume;			// Volume that this sound is playing at
	int				pan;			// Pan value that this sound is playing at
	short			handle;			// What handle this sound is playing on.  Valid only if SOF_PLAYING is set.
	short			soundnum;		// The sound number that is playing
	union {
		struct {
			short			segnum;	// Used if SOF_LINK_TO_POS field is used
			short			sidenum;				
			vms_vector		position;
		} pos;
		struct {
			short			objnum;	// Used if SOF_LINK_TO_OBJ field is used
			short			objsignature;
		} obj;
	} link_type;
} sound_object;

#define MAX_SOUND_OBJECTS 5 /* WHS: 16 in pc case */
sound_object SoundObjects[MAX_SOUND_OBJECTS];
short next_signature=0;

#define SOUND_OFFSET		10000

//short digi_sound_list[MAX_SOUNDS];

#ifdef MAC_SHAREWARE
static short sound_list[] = {
	10000,10001,10002,10003,10004,10005,10006,10007,10008,10009,10010,10011,10012,10013,10014,10015,10016,
	10017,10018,10019,10020,10021,10022,10023,10024,10025,10026,10027,10028,10029,10030,10031,10032,10033,
	10034,10035,10036,10037,10038,10039,10040,10041,10042,10043,10044,10045,10046,10047,10048,10049,10050,
	10051,10052,10053,10054,10055,10056,10057,10058,10059,100100,10061,10062,10063,10064,10065,10066,10067,
	10068,10069,10070,10071,10072,SOUND_END
};
#else
static short sound_list[] = {
	10000,10001,10002,10003,10004,10005,10006,10007,10008,10009,10010,10011,10012,10013,10014,10015,10016,
	10017,10018,10019,10020,10021,10022,10023,10024,10025,10026,10027,10028,10029,10030,10031,10032,10033,
	10034,10035,10036,10037,10038,10039,10040,10041,10042,10043,10044,10045,10046,10047,10048,10049,10050,
	10051,10052,10053,10054,10055,10056,10057,10058,10059,10060,10061,10062,10063,10064,10065,10066,10067,
	10068,10069,10070,10071,10072,10073,10074,10075,10076,10077,10078,10079,10080,10081,10082,10083,10084,
	10085,10086,10087,10088,10089,10090,10091,10092,10093,10094,10095,10096,10097, SOUND_END
};
#endif


void digi_reset_digi_sounds()
{
#ifdef USE_SOUND
	int i;

	DEB1(printf("X: digi_reset_digi_sound\n"));
	if ( !digi_initialized ) return;

		/* EndAllSound(); */
	gsi_kill_sounds();
#endif /* USE_SOUND */
}


int digi_xlat_sound( int soundno )
{
	DEB1(printf("X: digi_xlat_sound %d=%d\n", soundno, (int)(Sounds[soundno] + SOUND_OFFSET)));
	if ( soundno < 0 ) return -1;

	if ( use_sounds == 1 )	{
		soundno = AltSounds[soundno];
		if ( soundno == 255 ) return -1;
	}
//	return (int)(digi_sound_list[Sounds[soundno]] + SOUND_OFFSET);
	if (Sounds[soundno] == 255)
		return -1;

	return (int)(Sounds[soundno] + SOUND_OFFSET);
}


void digi_play_sample_3d( int sndnum, int angle, int volume, int no_dups )
{
#ifdef USE_SOUND
	int vol;
	int i = sndnum, demo_angle;
	DEB1(printf("X: digi_play_sample_3d (num %d)\n", sndnum));
	if (Newdemo_state == ND_STATE_RECORDING) {
		demo_angle = fixmuldiv(angle, F1_0, 255);
		if (no_dups)
			newdemo_record_sound_3d_once(sndnum, demo_angle, volume);
		else
			newdemo_record_sound_3d(sndnum, demo_angle, volume);
	}
	if (!digi_initialized) return;
	if (digi_paused) {
		digi_resume_all();
		if (digi_paused)
			return;
	}
	if ( sndnum < 0 ) return;
	i = digi_xlat_sound(sndnum);
	if (i == -1) return;

	vol = fixmuldiv(volume, digi_volume, F1_0);
//	ChangeSoundVolume( i, vol );
//	BeginSound(i, SOUND_RATE_11k);
//	ChangeSoundStereoPosition(i, angle);

	/* WHS: New GSI code */
	DEB1(printf("   vol=%d, angle=%d\n", vol, angle));
//angle =-256 - 0 - 256 =pan? i.e. angle =0-128, yep but mirrored!
	gsi_play_sound(i-10000, SND_PLAY);
	gsi_send_commands(4, /*SND_SET_ANGLE_DEFAULT_R*/ SND_SET_ANGLE | SND_CURRENT, ((angle+256)>>2)^127, SND_SET_VOLUME | SND_CURRENT, (vol>0 ? vol-1:0) );
//	gsi_send_commands(4, SND_SET_PAN | SND_CURRENT, angle, SND_SET_VOLUME | SND_CURRENT, (vol>0 ? vol-1:0) );

	gsi_flush();
#endif /* USE_SOUND */
}


void digi_play_sample( int sndnum, fix max_volume )
{
#ifdef USE_SOUND
	int i, vol;

	DEB1(printf("X: digi_play_sample (%d)\n", sndnum));
	if (Newdemo_state == ND_STATE_RECORDING)
		newdemo_record_sound(sndnum);

	if (!digi_initialized) return;
	if (digi_paused) {
		digi_resume_all();
		if (digi_paused)
			return;
	}
	if ( sndnum < 0 ) return;

	i = digi_xlat_sound(sndnum);
	if (i == -1) return;

	vol = fixmuldiv(max_volume, digi_volume, F1_0);
//	ChangeSoundVolume(i, vol );
//	err = BeginSound(i, SOUND_RATE_11k);
//	ChangeSoundStereoPosition(i, 0);

	/* WHS: New GSI code */
	DEB1(printf("   vol=%d\n", vol));
	gsi_play_sound(i-10000, SND_PLAY);
//gsi_send_commands(4, /*SND_SET_ANGLE_DEFAULT_R*/SND_SET_ANGLE  | SND_CURRENT, 0, SND_SET_VOLUME | SND_CURRENT, (vol>0? vol-1: 0));
	gsi_send_commands(2, SND_SET_VOLUME | SND_CURRENT, (vol>0? vol-1: 0));
	gsi_flush();
#endif /* USE_SOUND */
}


void digi_play_sample_once( int sndnum, fix max_volume )
{
#ifdef USE_SOUND
	DEB1(printf("X: digi_play_sample_once (%d)\n", sndnum));
	digi_play_sample( sndnum, F1_0 );
#endif /* USE_SOUND */
}


void digi_set_digi_volume(int dvolume)
{
#ifdef USE_SOUND
	int i;
	DEB1(printf("X: digi_set_volume (to %d)\n", dvolume));
	if ( !digi_initialized ) return;

//	dvolume = fixmuldiv( dvolume, _DIGI_MAX_VOLUME, 0x7fff);
	if ( dvolume > _DIGI_MAX_VOLUME )
		digi_volume = _DIGI_MAX_VOLUME;
	else if ( dvolume < 0 )
		digi_volume = 0;
	else
		digi_volume = dvolume;

	digi_sync_sounds();
#endif /* USE_SOUND */
}


void digi_stop_current_song()
{
#if 0 /* Midi not yet handled */
#ifdef USE_SOUND

#ifdef MAC_SHAREWARE
	if ( digi_midi_song_playing ) {
		if ( !IsSongDone() )
			EndSong();					// stop the song 
		FreeSong();						// free the resources used.
		ReleaseResource(Midi_handle);
		ReleaseResource(Song_handle);
		Song_handle = NULL;
		Midi_handle = NULL;
		digi_midi_song_playing = 0;
	}
#endif
#ifdef MAC_REGISTERED_15
	stop_redbook();
	digi_midi_song_playing = 0;
#endif
#endif /* USE_SOUND */
#endif /* not yet handled */
}


void digi_play_midi_song(int songnum, int loop )
{
#if 0 /* Midi not yet handled */
#ifdef USE_SOUND

#ifdef MAC_SHAREWARE
	song_resource *song;
	ushort midi_id;
#endif
#ifdef PC_REGISTERED_15
	char fname[128];
	CFILE		*fp;

	char *filename;
	char *melodic_bank;
	char *drum_bank;
#endif

	if (!digi_initialized) return;

	digi_last_midi_song = songnum;
	digi_last_midi_song_loop = loop;

#ifdef MAC_REGISTERED_15
	if ( !have_descent_cd ) {
		if ( !cd_playing && !digi_cd_started )
			redbook_start_play();
		digi_cd_started = 1;
		return;
	}
#endif

	digi_stop_current_song();

	if (midi_volume < 1)
		return;
#ifdef PC_REGISTERED_15
/* Use midi, grab the melodic stuff from the pc version's digi.c */
/* also grab the info for which banks should be played etc. */

	if ( digi_midi_type <= 0 )	return; /* No midi playing hardware or disabled */

	digi_stop_current_song();

	filename = Songs[songnum].filename;
	melodic_bank = Songs[songnum].melodic_bank_file;
	drum_bank = Songs[songnum].drum_bank_file;

	if ( filename == NULL )	return;

//	strcpy( digi_last_midi_song, filename );
//	strcpy( digi_last_melodic_bank, melodic_bank );
//	strcpy( digi_last_drum_bank, drum_bank );

	fp = NULL;
#if 0
	if ( (digi_midi_type==_MIDI_FM)||(digi_midi_type==_MIDI_OPL3) )	{	
		int sl;
		sl = strlen( filename );
		strcpy( fname, filename );	
		fname[sl-1] = 'q';
		fp = cfopen( fname, "rb" );
	}
#endif

	if ( !fp  )	{
		fp = cfopen( filename, "rb" );
		if (!fp) {
	 		mprintf( (1, "Error opening midi file, '%s'", filename ));
	 		return;
		}
	}
	if ( midi_volume < 1 )		{
		cfclose(fp);
		return;				// Don't play song if volume == 0;
	}
	SongSize = cfilelength( fp );
	SongData = malloc( SongSize );
	if (SongData==NULL)	{
		cfclose(fp);
		mprintf( (1, "Error mallocing %d bytes for '%s'", SongSize, filename ));
		return;
	}
	if ( cfread (  SongData, SongSize, 1, fp )!=1)	{
		mprintf( (1, "Error reading midi file, '%s'", filename ));
		cfclose(fp);
		free(SongData);
		SongData=NULL;
		return;
	}
	cfclose(fp);

#if 0
	if ( (digi_midi_type==_MIDI_FM)||(digi_midi_type==_MIDI_OPL3) )	{	
		if ( !digi_load_fm_banks(melodic_bank, drum_bank) )	{
			return;
		}
	}
#endif

//	if (!dpmi_lock_region(SongData, SongSize))	{
//		mprintf( (1, "Error locking midi file, '%s'", filename ));
//		free(SongData);
//		SongData=NULL;
//		return;
//	}

	// setup the song initialization structure
//	sSOSInitSong.lpSongData = SongData;
//	if ( loop )
//		sSOSInitSong.lpSongCallback = sosMIDICallback;
//	else
//		sSOSInitSong.lpSongCallback = _NULL;

//	for ( i=0; i<32; i++ )
//		sSOSTrackMap.wTrackDevice[i] = _MIDI_MAP_TRACK;
			
//	for ( i=0; i<_SOS_MIDI_MAX_TRACKS; i++ )
//		_lpSOSMIDITrack[0][i] = _NULL;

	//initialize the song
//	if( ( wError = sosMIDIInitSong( &sSOSInitSong, &sSOSTrackMap, &wSongHandle ) ) )	{
//		mprintf( (1, "\nHMI Error : %s", sosGetErrorString( wError ) ));
//		free(SongData);
//		SongData=NULL;
//		return;
//	}

//	Assert( wSongHandle == 0 );

	// start the song playing
//	if( ( wError = sosMIDIStartSong( wSongHandle ) ) ) {
//		mprintf( (1, "\nHMI Error : %s", sosGetErrorString( wError ) ));
//		// uninitialize the last MIDI song
//		sosMIDIUnInitSong( wSongHandle );
//		free(SongData);
//		SongData=NULL;
//		return;
//   }



#endif
#ifdef MAC_SHAREWARE /* This must be altered */
	Song_handle = GetResource('SONG', songnum);
	if (!Song_handle)
		return;
	song = (song_resource *)(*Song_handle);
	midi_id = song->midi_id;
	Midi_handle = GetResource('Midi', midi_id);
	if (!Midi_handle) {
		ReleaseResource(Song_handle);
		return;
	}
	song->sound_voices = digi_max_channels;
	song->norm_voices = 4 + (digi_max_channels/2);
	err = BeginSongFromMemory(songnum, Song_handle, Midi_handle, loop);
	if (err != noErr)
		return;
#endif
#ifdef MAC_REGISTERED_15
	play_redbook_track(songnum, loop);
#endif
	digi_midi_song_playing = 1;
#endif /* USE_SOUND */	
#endif /* not yet handled */
}


void digi_set_midi_volume(int n)
{
#if 0 /* Midi not yet handled */
#ifdef USE_SOUND
	int old_volume = midi_volume;

	if (!digi_initialized) return;

	if (n < 0)
		midi_volume = 0;
	else if (n > 255)
		midi_volume = 255;
	else
		midi_volume = n;

#ifdef MAC_SHAREWARE
	SetSongVolume(midi_volume);
#endif
#ifdef MAC_REGISTERED_15
	redbook_set_volume(midi_volume);
#endif

	if (!old_volume && midi_volume) {
#ifdef MAC_SHAREWARE
		ChangeSystemVoices(MAX_MUSIC_CHANNELS, 4 + (digi_max_channels/2), digi_max_channels);
#endif
		digi_play_midi_song(digi_last_midi_song, digi_last_midi_song_loop);
	} else if (old_volume && !midi_volume) {
		digi_stop_current_song();
#ifdef MAC_REGISTERED_15
		digi_cd_started = 0;
#endif
#ifdef MAC_SHAREWARE
		ChangeSystemVoices(0, digi_max_channels/2, digi_max_channels);
#endif

	}
#endif /* USE_SOUND */
#endif /* not yet handled */
}

void digi_set_volume(int dvolume, int mvolume)
{
#ifdef USE_SOUND
	DEB1(printf("X: digi_set_volume (d: %d, m: %d)\n", dvolume, mvolume));
	if (!digi_initialized) return;
	digi_set_digi_volume(dvolume);
	digi_set_midi_volume(mvolume);
#endif /* USE_SOUND */
}

void digi_set_master_volume( int volume )
{
#ifdef USE_SOUND
	DEB1(printf("X: digi_set_master_volume (to %d)\n", volume));

	if (!digi_initialized) return;

	Config_master_volume = volume;	
	if ( Config_master_volume > 8 )
		Config_master_volume = 8;
	else if ( Config_master_volume < 0 )
		Config_master_volume = 0;
		
//	SetMasterVolume( (Config_master_volume*256)/8 );
	gsi_set_master_volume( (Config_master_volume*256)/8 );
#endif /* USE_SOUND */
}

void digi_reset()
{
#ifdef USE_SOUND

	DEB1(printf("X: digi_reset\n"));

/*  from PC code:
	if ( Digi_initialized )	{
		digi_reset_digi_sounds();
		digi_close();
		mprintf( (0, "Sound system DISABLED.\n" ));
	} else {
		digi_init();
		mprintf( (0, "Sound system ENABLED.\n" ));
	}
*/
#endif /* USE_SOUND */
}

void digi_resume_all()
{
#ifdef USE_SOUND
	DEB1(printf("X: digi_resume_all\n"));

	if (!digi_initialized) return;

	if (digi_paused) {
		gsi_send_commands(1, SND_CONTINUE_ALL);
		gsi_flush();

#if 0 /* Music not yet handled */
#ifdef MAC_SHAREWARE
		ResumeMusicOnly();
#endif
#ifdef MAC_REGISTERED_15
		redbook_pause(0);
#endif
#endif /* not yet handled */
		digi_paused = 0;
	}
#endif /* USE_SOUND */
}

void digi_pause_all()
{
#ifdef USE_SOUND
	int i;

	DEB1(printf("X: digi_pause_all\n"));

	if (!digi_initialized) return;

	if (!digi_paused) {	
		gsi_send_commands(1, SND_PAUSE_ALL);
		gsi_flush();

		for (i=0; i<MAX_SOUND_OBJECTS; i++ )	{
			if ( (SoundObjects[i].flags & SOF_USED )	&& (SoundObjects[i].flags & SOF_PLAYING) && (SoundObjects[i].flags && SOF_PLAY_FOREVER) ) {
//				EndSound(SoundObjects[i].soundnum);
				SoundObjects[i].flags &= ~SOF_PLAYING;		// Mark sound as not playing
			}
		}
#if 0 /* Music not yet done */
#ifdef MAC_SHAREWARE
		PauseMusicOnly();
#endif
#ifdef MAC_REGISTERED_15
		redbook_pause(1);
#endif
#endif /* not yet handled */
		digi_paused = 1;
	}
#endif /* USE_SOUND */
}

void digi_stop_all()
{
#ifdef USE_SOUND
	int i;

	DEB1(printf("X: digi_stop_all\n"));
	if (!digi_initialized)	return;
//	EndAllSound();
	gsi_kill_sounds();
//	gsi_kill_song();

// reset the sound objects to be unused.  We stopped the sounds above.

	for (i = 0; i < MAX_SOUND_OBJECTS; i++)
		SoundObjects[i].flags = 0;
#endif /* USE_SOUND */
}

void digi_set_max_channels(int n)
{
#ifdef USE_SOUND
	song_resource *song;
	DEB1(printf("X: digi_set_max_channels (%d)\n", n));
	digi_max_channels	= n;

	if ( digi_max_channels < 1 ) 
		digi_max_channels = 1;
	if ( digi_max_channels > MAX_SOUND_CHANNELS ) 
		digi_max_channels = MAX_SOUND_CHANNELS;

	if ( !digi_initialized ) return;

#if 0 /* Midi not yet handled */
	if (digi_midi_song_playing)
		digi_stop_current_song();

	if (midi_volume <= 0)
		ChangeSystemVoices(0, digi_max_channels/2, digi_max_channels);
	else
		ChangeSystemVoices(MAX_MUSIC_CHANNELS, 4 + (digi_max_channels/2), digi_max_channels);
	
	if (digi_last_midi_song != 0)
		digi_play_midi_song(digi_last_midi_song, digi_last_midi_song_loop);
#endif /* } not yet handled */
	digi_reset_digi_sounds();
#endif /* } USE_SOUND */
}

void digi_start_sound_object(int i)
{
#ifdef USE_SOUND /* { */
	int vol;
	DEB1(printf("X: digi_start_sound_object %d\n", i));
	if (!digi_initialized) return;
	
	vol = fixmuldiv(SoundObjects[i].volume, digi_volume, F1_0);
//	ChangeSoundVolume(SoundObjects[i].soundnum, fixmuldiv(SoundObjects[i].volume, digi_volume, F1_0));
//	BeginSoundLoop( SoundObjects[i].soundnum, SOUND_RATE_11k, 1, GetSoundLength(SoundObjects[i].soundnum) );
//	ChangeSoundStereoPosition(SoundObjects[i].soundnum, SoundObjects[i].pan);

/* WHS: New GSI code */
	DEB1(printf("   vol=%d\n", vol));
	DEB1(printf("   SoundObjects[i].soundnum=%d\n", SoundObjects[i].soundnum));
	gsi_play_channel_sound(i, SoundObjects[i].soundnum-10000, SND_PLAY);
	gsi_send_commands(4, /*SND_SET_CHANNEL_ANGLE_DEFAULT_R*/SND_SET_CHANNEL_ANGLE | SND_CURRENT, SoundObjects[i].pan, SND_SET_CHANNEL_VOLUME | SND_CURRENT, /*vol>>8*/ (vol>0? vol-1: 0));
	gsi_flush();

	SoundObjects[i].flags |= SOF_PLAYING;
#endif /* USE_SOUND */
}

void digi_get_sound_loc( vms_matrix * listener, vms_vector * listener_pos, int listener_seg, vms_vector * sound_pos, int sound_seg, fix max_volume, int *volume, int *pan, fix max_distance )	
{
#ifdef USE_SOUND
	vms_vector	vector_to_sound;
	fix angle_from_ear, cosang,sinang;
	fix distance;
	fix path_distance;

	DEB1(printf("X: digi_get_sound_loc\n"));
	*volume = 0;
	*pan = 0;

	if (!digi_initialized) return;

	max_distance = (max_distance*5)/4;		// Make all sounds travel 1.25 times as far.

	//	Warning: Made the vm_vec_normalized_dir be vm_vec_normalized_dir_quick and got illegal values to acos in the fang computation.
	distance = vm_vec_normalized_dir_quick( &vector_to_sound, sound_pos, listener_pos );

	if (distance < max_distance )	{
		int num_search_segs = f2i(max_distance/20);
		if ( num_search_segs < 1 ) num_search_segs = 1;

		path_distance = find_connected_distance(listener_pos, listener_seg, sound_pos, sound_seg, num_search_segs, WID_RENDPAST_FLAG );
		//path_distance = distance;
		if ( path_distance > -1 )	{
			*volume = max_volume - (path_distance/f2i(max_distance));
			//mprintf( (0, "Sound path distance %.2f, volume is %d / %d\n", f2fl(distance), *volume, max_volume ));
			if (*volume > 0 )	{
				angle_from_ear = vm_vec_delta_ang_norm(&listener->rvec,&vector_to_sound,&listener->uvec);
				fix_sincos(angle_from_ear,&sinang,&cosang);
				//mprintf( (0, "volume is %.2f\n", f2fl(*volume) ));
				if (Config_channels_reversed) cosang *= -1;
				*pan = fixmuldiv(cosang, 255, F1_0);
			} else {
				*volume = 0;
			}
		}
	}
#endif /* USE_SOUND */
}

int digi_link_sound_to_object( int soundnum, short objnum, int forever, fix max_volume )
{
#ifdef USE_SOUND
	DEB1(printf("X: digi_link_sound_to_object (sndnum %d)\n", soundnum));
    // 10 segs away
	return digi_link_sound_to_object2( soundnum, objnum, forever, max_volume, 256*F1_0  );
#else
	return 0;
#endif /* USE_SOUND */
}

int digi_link_sound_to_pos( int soundnum, short segnum, short sidenum, vms_vector * pos, int forever, fix max_volume )
{
#ifdef USE_SOUND
	DEB1(printf("X: digi_link_sound_to_pos (sndnum %d)\n", soundnum));
	return digi_link_sound_to_pos2( soundnum, segnum, sidenum, pos, forever, max_volume, F1_0 * 256 );
#else
	return 0;
#endif /* USE_SOUND */
}

int digi_link_sound_to_object2( int org_soundnum, short objnum, int forever, fix max_volume, fix  max_distance )
{
#ifdef USE_SOUND
	int i,volume,pan;
	object * objp;
	int soundnum;

	DEB1(printf("X: digi_link_sound_to_object2 (sndnum %d)\n", org_soundnum));

	soundnum = digi_xlat_sound(org_soundnum);

	if (!digi_initialized) return -1;
	if ( max_volume < 0 ) return -1;
	if (soundnum < 0 ) return -1;
	if ((objnum<0)||(objnum>Highest_object_index))
		return -1;

	if ( !forever )	{
		// Hack to keep sounds from building up...
		digi_get_sound_loc( &Viewer->orient, &Viewer->pos, Viewer->segnum, &Objects[objnum].pos, Objects[objnum].segnum, max_volume,&volume, &pan, max_distance );
		digi_play_sample_3d( org_soundnum, pan, volume, 0 );
		return -1;
	}

	for (i=0; i<MAX_SOUND_OBJECTS; i++ )
		if (SoundObjects[i].flags==0)
			break;

	if (i==MAX_SOUND_OBJECTS) {
		mprintf((1, "Too many sound objects!\n" ));
		return -1;
	}

	SoundObjects[i].signature=next_signature++;
	SoundObjects[i].flags = SOF_USED | SOF_LINK_TO_OBJ;
	if ( forever )
		SoundObjects[i].flags |= SOF_PLAY_FOREVER;
	SoundObjects[i].link_type.obj.objnum = objnum;
	SoundObjects[i].link_type.obj.objsignature = Objects[objnum].signature;
	SoundObjects[i].max_volume = max_volume;
	SoundObjects[i].max_distance = max_distance;
	SoundObjects[i].volume = 0;
	SoundObjects[i].pan = 0;
	SoundObjects[i].soundnum = (short)soundnum;

	objp = &Objects[SoundObjects[i].link_type.obj.objnum];
	digi_get_sound_loc( &Viewer->orient, &Viewer->pos, Viewer->segnum, 
                       &objp->pos, objp->segnum, SoundObjects[i].max_volume,
                       &SoundObjects[i].volume, &SoundObjects[i].pan, SoundObjects[i].max_distance );

	digi_start_sound_object(i);

	return SoundObjects[i].signature;
#else
	return 0;
#endif /* USE_SOUND */
}

int digi_link_sound_to_pos2( int org_soundnum, short segnum, short sidenum, vms_vector * pos, int forever, fix max_volume, fix max_distance )
{
#ifdef USE_SOUND
	int i, volume, pan;
	int soundnum;

	DEB1(printf("X: digi_link_sound_to_pos2 (sndnum %d)\n", org_soundnum));

	soundnum = digi_xlat_sound(org_soundnum);

	if (!digi_initialized) return -1;
	if ( max_volume < 0 ) return -1;

	if (soundnum < 0 ) return -1;
	if ((segnum<0)||(segnum>Highest_segment_index))
		return -1;

	if ( !forever )	{
		// Hack to keep sounds from building up...
		digi_get_sound_loc( &Viewer->orient, &Viewer->pos, Viewer->segnum, pos, segnum, max_volume, &volume, &pan, max_distance );
		digi_play_sample_3d( org_soundnum, pan, volume, 0 );
		return -1;
	}

	for (i=0; i<MAX_SOUND_OBJECTS; i++ )
		if (SoundObjects[i].flags==0)
			break;

	if (i==MAX_SOUND_OBJECTS) {
		mprintf((1, "Too many sound objects!\n" ));
		return -1;
	}

	SoundObjects[i].signature=next_signature++;
	SoundObjects[i].flags = SOF_USED | SOF_LINK_TO_POS;
	if ( forever )
		SoundObjects[i].flags |= SOF_PLAY_FOREVER;
	SoundObjects[i].link_type.pos.segnum = segnum;
	SoundObjects[i].link_type.pos.sidenum = sidenum;
	SoundObjects[i].link_type.pos.position = *pos;
	SoundObjects[i].soundnum = (short)soundnum;
	SoundObjects[i].max_volume = max_volume;
	SoundObjects[i].max_distance = max_distance;
	SoundObjects[i].volume = 0;
	SoundObjects[i].pan = 0;
	digi_get_sound_loc( &Viewer->orient, &Viewer->pos, Viewer->segnum, 
                       &SoundObjects[i].link_type.pos.position, SoundObjects[i].link_type.pos.segnum, SoundObjects[i].max_volume,
                       &SoundObjects[i].volume, &SoundObjects[i].pan, SoundObjects[i].max_distance );

	digi_start_sound_object(i);

	return SoundObjects[i].signature;
#else
	return 0;
#endif /* USE_SOUND */
}

void digi_sync_sounds()
{
#ifdef USE_SOUND
	int i;
	int oldvolume, oldpan;

	DEB1(printf("X: digi_sync_sounds\n"));

	if (!digi_initialized) return;

	for (i=0; i<MAX_SOUND_OBJECTS; i++ )	{
		if ( SoundObjects[i].flags & SOF_USED )	{
			oldvolume = SoundObjects[i].volume;
			oldpan = SoundObjects[i].pan;

			if ( !(SoundObjects[i].flags & SOF_PLAY_FOREVER) )	{
				// Check if its done.
				if (SoundObjects[i].flags & SOF_PLAYING) {
//					if ( IsThisSoundFXFinished(SoundObjects[i].soundnum) ) {
if (gsi_get_channel_status(i /*SoundObjects[i].soundnum-10000*/)==0) {
						SoundObjects[i].flags = 0;	// Mark as dead, so some other sound can use this sound
						continue;		// Go on to next sound...
					}
				}
			}

			if ( SoundObjects[i].flags & SOF_LINK_TO_POS )	{
				digi_get_sound_loc( &Viewer->orient, &Viewer->pos, Viewer->segnum, 
                                &SoundObjects[i].link_type.pos.position, SoundObjects[i].link_type.pos.segnum, SoundObjects[i].max_volume,
                                &SoundObjects[i].volume, &SoundObjects[i].pan, SoundObjects[i].max_distance );

			} else if ( SoundObjects[i].flags & SOF_LINK_TO_OBJ )	{
				object * objp;

				objp = &Objects[SoundObjects[i].link_type.obj.objnum];

				if ((objp->type==OBJ_NONE) || (objp->signature!=SoundObjects[i].link_type.obj.objsignature))	{
					// The object that this is linked to is dead, so just end this sound if it is looping.
					if ( (SoundObjects[i].flags & SOF_PLAYING)  && (SoundObjects[i].flags & SOF_PLAY_FOREVER))	{
//						EndSound(SoundObjects[i].soundnum);
						DEB1(printf("SoundObjects[i].soundnum=%d\n", SoundObjects[i].soundnum));
//						gsi_kill_sound(SoundObjects[i].soundnum-10000);
						gsi_kill_channel_sound(i);
					}
					SoundObjects[i].flags = 0;	// Mark as dead, so some other sound can use this sound
					continue;		// Go on to next sound...
				} else {
					digi_get_sound_loc( &Viewer->orient, &Viewer->pos, Viewer->segnum, 
	                                &objp->pos, objp->segnum, SoundObjects[i].max_volume,
                                   &SoundObjects[i].volume, &SoundObjects[i].pan, SoundObjects[i].max_distance );
				}
			}

			if (oldvolume != SoundObjects[i].volume) 	{
				if ( SoundObjects[i].volume < 1 )	{
					// Sound is too far away, so stop it from playing.
					if ((SoundObjects[i].flags & SOF_PLAYING)&&(SoundObjects[i].flags & SOF_PLAY_FOREVER))	{
//						EndSound(SoundObjects[i].soundnum);
						DEB1(printf("SoundObjects[i].soundnum=%d\n", SoundObjects[i].soundnum));
//						gsi_kill_sound(SoundObjects[i].soundnum-10000);
						gsi_kill_channel_sound(i);

						SoundObjects[i].flags &= ~SOF_PLAYING;		// Mark sound as not playing
					}
				} else {
					if (!(SoundObjects[i].flags & SOF_PLAYING))	{
						digi_start_sound_object(i);
					} else {
						int vol;

						vol = fixmuldiv(SoundObjects[i].volume, digi_volume,F1_0);
//						ChangeSoundVolume(SoundObjects[i].soundnum, fixmuldiv(SoundObjects[i].volume,digi_volume,F1_0) );
//						gsi_set_volume(0, SND_CURRENT, fixmuldiv(SoundObjects[i].volume,digi_volume,F1_0) );
						gsi_send_commands(3, SND_SET_CHANNEL_VOLUME, i, max(fixmuldiv(SoundObjects[i].volume,digi_volume,F1_0)-1,0));
					}
				}
			}

			if (oldpan != SoundObjects[i].pan) 	{
				if (SoundObjects[i].flags & SOF_PLAYING) {
					int x=SoundObjects[i].pan;
//					ChangeSoundStereoPosition( SoundObjects[i].soundnum, SoundObjects[i].pan );
					DEB1(printf("pan=%d\n", SoundObjects[i].pan));
					gsi_send_commands(3, /*SND_SET_CHANNEL_ANGLE_DEFAULT_R*/ SND_SET_CHANNEL_ANGLE, i, SoundObjects[i].pan);
				}
			}
		}
	}
	gsi_flush();
#endif /* USE_SOUND */
}

void digi_kill_sound_linked_to_segment( int segnum, int sidenum, int soundnum )
{
#ifdef USE_SOUND
	int i,killed;

	DEB1(printf("X: digi_kill_sound_linked_to_segment (sndnum %d)\n", soundnum));

	soundnum = digi_xlat_sound(soundnum);

	if (!digi_initialized) return;

	killed = 0;

	for (i=0; i<MAX_SOUND_OBJECTS; i++ )	{
		if ( (SoundObjects[i].flags & SOF_USED) && (SoundObjects[i].flags & SOF_LINK_TO_POS) )	{
			if ((SoundObjects[i].link_type.pos.segnum == segnum) && (SoundObjects[i].soundnum==(short)soundnum ) && (SoundObjects[i].link_type.pos.sidenum==sidenum) )	{
				if ( SoundObjects[i].flags & SOF_PLAYING )	{
//					EndSound(SoundObjects[i].soundnum);
					DEB1(printf("SoundObjects[i].soundnum=%d\n", SoundObjects[i].soundnum));
//					gsi_kill_sound(SoundObjects[i].soundnum-10000);
					gsi_kill_channel_sound(i);
				}
				SoundObjects[i].flags = 0;	// Mark as dead, so some other sound can use this sound
				killed++;
			}
		}
	}
	// If this assert happens, it means that there were 2 sounds
	// that got deleted. Weird, get John.
	if ( killed > 1 )	{
		mprintf( (1, "ERROR: More than 1 sounds were deleted from seg %d\n", segnum ));
	}
	gsi_flush();
#endif /* USE_SOUND */
}

void digi_kill_sound_linked_to_object( int objnum )
{
#ifdef USE_SOUND
	int i,killed;

	DEB1(printf("X: digi_kill_sound_linked_to_object (objnum %d)\n", objnum));

	if (!digi_initialized) return;

	killed = 0;

	for (i=0; i<MAX_SOUND_OBJECTS; i++ )	{
		if ( (SoundObjects[i].flags & SOF_USED) && (SoundObjects[i].flags & SOF_LINK_TO_OBJ ) )	{
			if (SoundObjects[i].link_type.obj.objnum == objnum)	{
				if ( SoundObjects[i].flags & SOF_PLAYING )	{
//					EndSound(SoundObjects[i].soundnum);
					/* GSI code */
					DEB1(printf("SoundObjects[i].soundnum=%d\n", SoundObjects[i].soundnum));
//					gsi_kill_sound(SoundObjects[i].soundnum-10000);
					gsi_kill_channel_sound(i);

				}
				SoundObjects[i].flags = 0;	// Mark as dead, so some other sound can use this sound
				killed++;
			}
		}
	}
	// If this assert happens, it means that there were 2 sounds
	// that got deleted. Weird, get John.
	if ( killed > 1 )	{
		mprintf( (1, "ERROR: More than 1 sounds were deleted from object %d\n", objnum ));
	}
	gsi_flush();
#endif /* USE_SOUND */
}


void digi_close()
{
#ifdef USE_SOUND
	int i, sound_num;

	DEB1(printf("X: digi_close\n"));

	if (!digi_initialized)
		return;
//	digi_stop_current_song(); /* not yet done */
//	SetMasterVolume(master_save);
	gsi_set_master_volume(master_save);

// free the sounds that we have registered.

	for (i = 0; i < num_sounds; i++) {
		sound_num = digi_xlat_sound(i);
		if (sound_num != -1)
//			FreeSound(i);
	gsi_flush(); //WHS: temp nop!
	}
//	ReleaseRegisteredSounds();
//	FinisSoundMusicSystem();

	digi_initialized = 0;
#endif /* USE_SOUND */
}

void digi_init_sounds()
{
#ifdef USE_SOUND
	int i;

	DEB1(printf("X: digi_init_sounds\n"));

	if (!digi_initialized) return;
//	EndAllSound();			// kill off all current sound effect and music
	gsi_kill_sounds();
//	gsi_kill_songs();
#endif /* USE_SOUND */
}

void digi_load_sounds()
{
#ifdef USE_SOUND
	int i, sound_num;

	DEB1(printf("X: digi_load_sounds\n"));

	if (!digi_initialized) return;	

//	if ( RegisterSounds(sound_list, 1) )
//		digi_close();


// on low memory conditions, we need to release sounds from the registered
// set so the memory can be freed.  The call to RegisterSounds only registers
// them with halestorm.  We will use LoadSound to actually load those sounds that
// we intend to use.

	for (i = 0; i < MAX_SOUNDS; i++) {
		sound_num = digi_xlat_sound(i);
#if 1
		if (sound_num != -1)
//			LoadSound(sound_num);		// load the sound into memory..hales says this is locked high in heap
		   gsi_flush();
#endif
	}
#endif /* USE_SOUND */
}

int digi_init()
{
#ifdef USE_SOUND
	CFILE *fp;
	int i;

	DEB1(printf("X: digi_init\n"));

	if (!use_sounds)
		return -1;

	num_sounds = (sizeof(sound_list) / sizeof(short)) - 1;

	gsi_init();

	//speed/channels/dsp mode/fragsize/sounddir/sound list
	gsi_init_pcm(11025, 2 /*stereo*/, GSI_8BIT,  8+2/*snd_channels*/-1+(2/*snd_type*/+1)/2-1, NULL, NULL);

#if 0 /* Midi not yet handled */
	i = MaxVoiceLoad();
	err = InitSoundMusicSystem(MAX_MUSIC_CHANNELS, 4 + (digi_max_channels/2), digi_max_channels, digi_sound_quality);
//	err = InitSoundMusicSystem(MAX_MUSIC_CHANNELS, 10, 16, digi_sound_quality);
	if (err != noErr)
		return -1;
#endif /* not yet handled */

//	master_save = GetMasterVolume();
	master_save = gsi_get_master_volume();

	if (master_save < 0)
		master_save = 0;
	else if ( master_save > 255 )
		master_save = 255;

#ifdef MAC_SHAREWARE
	PurgeSongs(1);
	LockSongs(0);
#endif
#ifdef MAC_REGISTERED_15
	init_redbook();
#endif

 	digi_initialized = 1;

	// set volumes and then out of here..

	digi_set_master_volume(Config_master_volume);
	digi_set_volume(digi_volume, midi_volume);
	digi_reset_digi_sounds();
	atexit(digi_close);
	DEB1(printf("digi_init: digi_initialized = %d\n", digi_initialized ));
#endif /* USE_SOUND */
	return 0;
}
